# -*- coding: utf-8 -*-
"""
Created on 2020/8/5
@project: SPAIC
@filename: Network
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description:
定义网络以及子网络，网络包含所有的神经网络元素、如神经集群、连接以及学习算法、仿真器等，实现最终的网络仿真与学习。
执行过程：网络定义->网络生成->网络仿真与学习
"""
from spaic.Network.Assembly import Assembly
# from spaic.Network.Topology import Connection
from collections import OrderedDict
from warnings import warn
import spaic
from spaic.utils.memory import get_tens_mem
from multiprocessing.pool import ThreadPool as Pool

try:
    import torch
except:
    pass


class Network(Assembly):
    _class_label = '<net>'

    def __init__(self, name=None):

        super(Network, self).__init__(name=name)
        self._monitors = OrderedDict()
        self._learners = OrderedDict()
        self._pipline = None
        self._backend: spaic.Backend = None
        self._forward_build = False
        pass

    # --------- Frontend code ----------
    def set_backend(self, backend=None, device='cpu', partition=False):
        if isinstance(device, str):
            device = [device]
        if backend is None:
            self._backend = spaic.Torch_Backend(device)
            self._backend.partition = partition
        elif isinstance(backend, spaic.Backend):
            self._backend = backend
        elif isinstance(backend, str):
            if backend == 'torch' or backend == 'pytorch':
                self._backend = spaic.Torch_Backend(device)
                self._backend.partition = partition
            elif backend == 'tensorflow':
                self._backend = spaic.Tensorflow_Backend(device)

    def set_backend_dt(self, dt=0.1, partition=False):
        if self._backend is None:
            warn("have not set backend, default pytorch backend is set automatically")
            self._backend = spaic.Torch_Backend('cpu')
            self._backend.dt = dt
        else:
            self._backend.dt = dt
            self._backend.partition = partition

    def set_random_seed(self, seed):
        if isinstance(self._backend, spaic.Torch_Backend):
            import torch
            torch.random.manual_seed(int(seed))
            if self._backend.device == 'cuda':
                torch.cuda.manual_seed(int(seed))

    def get_testparams(self):
        self.all_Wparams = list()
        for key, value in self._backend._parameters_dict.items():
            self.all_Wparams.append(value)
        return self.all_Wparams

    def add_learner(self, name, learner):
        assert isinstance(learner, spaic.Learner)
        self.__setattr__(name, learner)

    # TODO: 这里的setattr是否有必要要，是否可以全部放到Assembly里？
    def __setattr__(self, name, value):
        from spaic.Monitor.Monitor import Monitor
        from spaic.Learning.Learner import Learner
        super(Network, self).__setattr__(name, value)
        if isinstance(value, Monitor):
            self._monitors[name] = value

        elif isinstance(value, Learner):
            self._learners[name] = value

    # ---------  backend code  ----------

    def build(self, backend=None, strategy=0, full_enable_grad=None, device=None):
        if full_enable_grad is not None:
            self.enable_full_grad(full_enable_grad)
        if self._backend is None:
            if backend is not None:
                if device is not None:
                    self.set_backend(backend, device)
                else:
                    self.set_backend(backend)
            else:
                if device is not None:
                    self.set_backend(device=device)
                else:
                    self.set_backend()

        self._backend.clear_step()

        # build 试运行时，假设一个runtime
        if self._backend.runtime is None:
            self._backend.runtime = 10 * self._backend.dt

        all_groups = self.get_groups()
        for asb in all_groups:
            asb.set_id()

        self.build_projections(self._backend)

        all_connections = self.get_connections()

        # for debug
        con_debug = False
        con_syn_count = 0

        for con in all_connections:
            con.set_id()

            # ----根据连接，对每个神经元建立input_connection和output_connection
            con.pre.register_connection(con, True)
            con.post.register_connection(con, False)

        if strategy == 1:
            # 采取单纯的从头递归地build，一旦出现环路会陷入死循环，可以避开固有延迟的问题,
            # Use directly build strategy to avoid inherent delay. But cannot be used on models with loop, will fall in an endless loop.
            # Unfortunately,
            self.forward_build(all_groups, all_connections)
            self._backend.forward_build = True
        # elif strategy == 2:
        #     # 采取策略性构建，但是目前存在两个问题：
        #     #   1. 网络中存在Assembly块时会出现bug，尚未修复
        #     #   2. Connection所使用的input_spike为上一步的，需要添加[updated]，目前暂不使用所以未添加
        #     #   该构建方式可以较大程度上避开固有延迟的问题
        #     self.strategy_build(self.get_groups(False))
        else:
            # 原本的构建方式，首先构建连接，每个连接都是用上一轮神经元的输出脉冲，从而存在固有延迟的问题
            # 但是可以避开环路的问题。
            self._backend.forward_build = False
            def build_fn(module):
                # if con_debug:
                #     con_syn_count += torch.count_nonzero(connection.weight.value).item()
                module.build(self._backend)

            # for connection in all_connections:
            #     connection.build(self._backend)
            #     if con_debug:
            #         import torch
            #         con_syn_count += torch.count_nonzero(connection.weight.value).item()

            # for group in all_groups:
            #     group.build(self._backend)    
            pool = Pool(4)
            pool.map(build_fn, all_connections)
            pool.close()
            pool.join()
            pool = Pool(4)
            pool.map(build_fn, all_groups)
            pool.close()
            pool.join()

        for learner in self._learners.values():
            learner.set_id()
            learner.build(self._backend)

        for monitor in self._monitors.values():
            monitor.build(self._backend)

        self._backend.build_graph()
        # self._backend.build()
        self._backend.builded = True
        # if con_debug:
        #     print("Connection synapses count:%d"%con_syn_count)

        # for group in all_groups:
        #     if hasattr(group, 'index'):
        #         group.index = 0

        pass

    def forward_build(self, all_groups=None, all_connections=None):
        builded_groups = []
        builded_connections = []
        nod_groups = []
        for group in all_groups.copy():
            if group._class_label == '<nod>':
                if (group._node_sub_class == '<encoder>') or (group._node_sub_class == '<generator>'):
                    group.build(self._backend)
                    builded_groups.append(group)
                    all_groups.remove(group)
                    for conn in group._output_connections:
                        self.deep_forward_build(conn, all_groups, all_connections, builded_groups, builded_connections)
                    for module in group._output_modules:
                        self.deep_forward_build(module, all_groups, all_connections, builded_groups,
                                                builded_connections)
                else:
                    all_groups.remove(group)
                    nod_groups.append(group)

        while all_groups or all_connections:
            for group in all_groups:
                self.deep_forward_build(group, all_groups, all_connections, builded_groups, builded_connections)
            for conn in all_connections:
                self.deep_forward_build(conn, all_groups, all_connections, builded_groups, builded_connections)

        for group in nod_groups:
            group.build(self._backend)
            builded_groups.append(group)

    def deep_forward_build(self, target, all_groups, all_connections, builded_groups, builded_connections):
        if (target in builded_groups) or (target in builded_connections):
            return
        if target._class_label == '<con>':
            pre = [target.pre]
            post = [target.post]
        elif target._class_label == '<neg>':
            pre = target._input_connections + target._input_modules
            post = target._output_connections + target._output_modules
        elif target._class_label == '<mod>':
            pre = target.input_targets.copy()
            post = target.output_targets.copy()
        else:
            raise ValueError("Deep forward build Error, unsupported class label.")

        for pr in pre:
            if (pr in all_groups) or (pr in all_connections):
                return

        target.build(self._backend)
        if target._class_label == '<con>':
            builded_connections.append(target)
            all_connections.remove(target)
        elif (target._class_label == '<neg>') or (target._class_label == '<mod>'):
            builded_groups.append(target)
            all_groups.remove(target)

        for po in post:
            self.deep_forward_build(po, all_groups, all_connections, builded_groups, builded_connections)

        return

    # def strategy_build(self, all_groups=None):
    #     builded_groups = []
    #     unbuild_groups = {}
    #     output_groups = []
    #     level = 0
    #     from spaic.Neuron.Node import Encoder, Decoder, Generator
    #     # ===================从input开始按深度构建计算图==============
    #     for group in all_groups:
    #         if isinstance(group, Encoder) or isinstance(group, Generator):
    #             # 如果是input节点，则开始深度构建计算图
    #             group.build(self._backend)
    #             builded_groups.append(group)
    #             # all_groups.remove(group)
    #             for conn in group._output_connections:
    #                 builded_groups, unbuild_groups = self.deep_build_conn(conn, builded_groups,
    #                                                                       unbuild_groups, level)
    #         elif isinstance(group, Decoder):
    #             # 如果节点是output节点，则放入output组在最后进行构建
    #             output_groups.append(group)
    #         else:
    #             if (not group._input_connections) and (not group._output_connections):
    #                 # 孤立点的情况
    #                 import warnings
    #                 warnings.warn('Isolated group occurs, please check the network.')
    #                 group.build(self._backend)
    #
    #     if unbuild_groups:
    #         import warnings
    #         warnings.warn('Loop occurs')
    #     # ====================开始构建环路==================
    #     for key in unbuild_groups.keys():
    #         for i in unbuild_groups[key]:
    #             if i in builded_groups:
    #                 continue
    #             else:
    #                 builded_groups = self.deep_build_neurongroup_with_delay(i, builded_groups)
    #
    #     # ====================构建output节点===============
    #     for group in output_groups:
    #         group.build(self._backend)
    #
    # def deep_build_neurongroup(self, neuron=None, builded_groups=None, unbuild_groups=None, level=0):
    #     conns = [i for i in neuron._input_connections if i not in builded_groups]
    #     # conns表示神经元还没有被建立的依赖连接
    #     if conns: #==========如果存在conns说明有input_connections还没有被build===========
    #         if str(level) in unbuild_groups.keys():
    #             unbuild_groups[str(level)].append(neuron)
    #         else:
    #             unbuild_groups[str(level)] = [neuron]
    #         return builded_groups, unbuild_groups
    #     else:
    #
    #         if neuron not in builded_groups:
    #             if neuron._class_label == '<asb>':
    #                 neuron.build(self._backend, strategy=2)
    #             else:
    #                 neuron.build(self._backend)
    #             builded_groups.append(neuron)
    #             for conn in neuron._output_connections:
    #                 builded_groups, unbuild_groups = self.deep_build_conn(conn, builded_groups,
    #                                                                       unbuild_groups, level)
    #         return builded_groups, unbuild_groups
    #
    # def deep_build_conn(self, conn=None, builded_groups=None, unbuild_groups=None, level=0):
    #     conn.build(self._backend)
    #     builded_groups.append(conn)
    #     level += 1
    #     builded_groups, unbuild_groups = self.deep_build_neurongroup(conn.post_assembly, builded_groups, unbuild_groups, level)
    #     return builded_groups, unbuild_groups
    #
    # def deep_build_conn_with_delay(self, conn, builded_groups):
    #     conn.build(self._backend)
    #     builded_groups.append(conn)
    #     if conn.post_assembly not in builded_groups:
    #         builded_groups = self.deep_build_neurongroup_with_delay(conn.post_assembly, builded_groups)
    #     return builded_groups
    #
    # def deep_build_neurongroup_with_delay(self, neuron, builded_groups):
    #     conns = [i for i in neuron._input_connections if i not in builded_groups]
    #     if conns:
    #         for conn in conns:
    #             conn.build(self._backend)
    #             builded_groups.append(conn)
    #         neuron.build(self._backend)
    #     else:
    #         neuron.build(self._backend)
    #     builded_groups.append(neuron)
    #     for conn in neuron._output_connections:
    #         if conn not in builded_groups:
    #             builded_groups = self.deep_build_conn_with_delay(conn, builded_groups)
    #     return builded_groups

    def run(self, backend_time):
        self._backend.set_runtime(backend_time)
        if self._backend.builded is False:
            self.build()
        self._backend.initial_step()
        self._backend.update_time_steps()

    def run_continue(self, backend_time):
        self._backend.set_runtime(backend_time)
        if self._backend.builded is False:
            self.build()
            self._backend.initial_step()
        self._backend.initial_continue_step()
        self._backend.update_time_steps()

    def reset(self, ):
        if self._backend.builded is True:
            self._backend.initial_step()

    def enable_full_grad(self, requires_grad=True):
        self._backend.full_enable_grad = requires_grad

    def init_run(self):
        self._backend.initial_step()

    def add_monitor(self, name, monitor):
        from spaic.Monitor.Monitor import Monitor
        assert isinstance(monitor, Monitor), "Type Error, it is not monitor"
        assert monitor not in self._monitors.values(), "monitor %s is already added" % (name)
        assert name not in self._monitors.keys(), "monitor with name: %s have the same name with an already exists monitor" % (
            name)

        self.__setattr__(name, monitor)
        # self._monitors[name] = monitor

    def get_elements(self):
        element_dict = dict()
        for element in self.get_groups():
            element_dict[element.id] = element
        return element_dict

    def save_state(self, filename=None, direct=None, save=True, hdf5=False):
        """
        Save weights in memory or on hard disk.

        Args:
            filename: The name of saved file.
            direct: Target direction for saving state.
            mode: Determines whether saved in hard disk, default set false, it means will not save on disk.

        Returns:
            state: Connections' weight of the network.

        """
        state = self._backend._parameters_dict
        if not save:
            return state
        if not filename:
            filename = self.name if self.name else 'autoname'
        if not direct:
            direct = './'
        file = filename.split('.')[0]
        path = direct + file + '/parameters/'
        import os
        import torch
        origin_path = os.getcwd()
        os.chdir(direct)
        if file not in os.listdir():
            os.mkdir(file)
        if 'parameters' not in os.listdir('./' + file):
            os.mkdir('./' + file + '/parameters')
            # os.mkdir('./NetData/' + dict + '/backend/_parameters_dict')
        os.chdir('./' + file + '/parameters')

        if hdf5:
            import h5py
            filename = filename if direct.endswith('.hdf5') else direct + '.hdf5'
            with h5py.File(direct, "w") as f:
                for i, item in enumerate(state):
                    f.create_dataset(item, data=self._backend._parameters_dict[item].cpu().detach().numpy())
                    # torch.save(self._backend._parameters_dict[item], os.getcwd()+'/'+str(i)+'.pt')
                    print(i, item, ': saved')
        else:
            torch.save(self._backend._parameters_dict, './_parameters_dict.pt')
            module_dict = {}
            module_exist = False
            for group in self.get_groups():
                if isinstance(group, spaic.Module):
                    module_dict[group.id] = group.state_dict
                    module_exist = True
            if module_exist:
                torch.save(module_dict, './module_dict.pt')
        os.chdir(origin_path)
        return

    def state_from_dict(self, state=False, filename=None, direct=None, device=None):
        """
        Reload states from memory or disk.

        Args:
            state: contains backend._parameters_dict .
            filename: The name of saved file.
            direct: Target direction for reloading state.
            mode: Determines whether saved in hard disk, default set false, it means will not save on disk.

        Returns:
            state: Connections' weight of the network.

        """
        if not self._backend:
            if device:
                self.set_backend('torch', device=device)
            else:
                self.set_backend('torch')
        if self._backend.builded is False:
            self.build()
        if self._backend.device != device:
            import warnings
            warnings.warn(
                'Backend device setting is ' + str(self._backend.device) + '. Backend device selection is priority.')
            # device = self._backend.device
        if state:
            import torch
            if isinstance(state, dict) or isinstance(state, torch.Tensor):
                for key, para in state.items():
                    backend_key = self._backend.check_key(key, self._backend._parameters_dict)
                    if backend_key:
                        target_device = self._backend._parameters_dict[backend_key].device
                        self._backend._parameters_dict[backend_key] = para.to(target_device)

                # if self._backend.device
                return
            else:
                raise ValueError("Given state has wrong type")

        if direct:
            if filename:
                path = direct + '/' + filename + '/parameters/'
            else:
                path = direct + '/parameters/'
        else:
            if filename:
                path = './' + filename + '/parameters/'
            else:
                path = './parameters/'

        import os
        import torch
        origin_path = os.getcwd()
        try:
            os.chdir(path)
        except:
            raise ValueError('Wrong Path.')

        if '_parameters_dict.pt' in os.listdir('./'):
            data = torch.load('./_parameters_dict.pt', map_location=self._backend.device0)
            for key, para in data.items():
                backend_key = self._backend.check_key(key, self._backend._parameters_dict)
                if backend_key:
                    target_device = self._backend._parameters_dict[backend_key].device
                    self._backend._parameters_dict[backend_key] = para.to(target_device)
            if 'module_dict.pt' in os.listdir('./'):
                module_data = torch.load('./module_dict.pt', map_location=self._backend.device0)
                for group in self.get_groups():
                    if isinstance(group, spaic.Module):
                        target_key = self._backend.check_key(group.id, module_data)
                        group.load_state_dict(module_data[target_key])
        else:
            for file in os.listdir('./'):
                if file.endswith('.hdf5'):
                    import h5py
                    with h5py.File(direct, 'r') as f:
                        for key, para in f.items():
                            backend_key = self._backend.check_key(key, self._backend._parameters_dict)
                            if key:
                                target_device = self._backend._parameters_dict[backend_key].device
                                self._backend._parameters_dict[backend_key] = para.to(target_device)
        os.chdir(origin_path)
        return



# -*- coding: utf-8 -*-
"""
Created on 2020/9/9
@project: SPAIC
@filename: BaseModule
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description: 
"""
from abc import abstractmethod
from collections import OrderedDict
import spaic
from typing import Optional,Any,List
from dataclasses import dataclass,field
from copy import copy
import uuid



class BaseModule():
    '''
    Base class for all snn modules (assemblies, connection, learner, monitor, piplines).

    '''
    _Module_Count = 0
    _class_label = '<bm>'

    def __init__(self):
        self.id = None
        self.name = None
        self.enabled = True
        self.training = True
        self._backend: spaic.Backend = None
        self._supers = []
        self._var_names = []
        self._var_dict = dict()
        self._ops = list()
        self.prefer_device = None

    @abstractmethod
    def build(self, backend):
        NotImplementedError()

    @abstractmethod
    def get_str(self, level):
        NotImplementedError()

    def set_name(self, given_name):

        if isinstance(given_name, str):
            if self.name is None:
                self.name = given_name
            elif 'autoname' in self.name:
                self.name = given_name

        # elif isinstance(given_name, list):
        #     context = given_name[-1]
        #     spaic.global_module_name_count += 1
        #     self.name = context.name +'subgroup' + str(spaic.global_module_name_count)

        else:
            spaic.global_module_name_count += 1
            self.name = 'autoname' + str(spaic.global_module_name_count)

        return self.name

    def set_id(self):
        if len(self._supers) == 0:
            self.id = self.name + self.__class__._class_label
        else:
            super_ids = []
            for super in self._supers:
                if super.id is not None:
                    super_ids.append(super.id)
                else:
                    super_ids.append(super.set_id())

            self.id = self.name + self.__class__._class_label
            if len(super_ids) == 1:
                self.id = super_ids[0] + '_' + self.id
            else:
                pre_id = '/'
                for prefix in super_ids:
                    pre_id += prefix + ','
                pre_id += '/'
                self.id = pre_id + '_' + self.id
        return self.id

    def set_build_level(self, level):

        if self.build_level < 0:
            self.build_level = level
        elif self.build_level > level:
            self.build_level = level

    def variable_to_backend(self, name, shape, value=None, is_parameter=False, is_sparse=False, init=None, init_param=None,
                            min=None, max=None, is_constant=False, prefer_device=None):
        self._var_names.append(name)
        self._var_dict[name] = self._backend.add_variable(self, name, shape, value, is_parameter, is_sparse, init, init_param, min, max, is_constant, prefer_device)
        return self._var_dict[name]

    def op_to_backend(self, outputs:list, func:callable, inputs:list):
        # check if the inputs and outputs variables belongs to this module object, if backend don't have this variable it will be added the module label
        if isinstance(inputs, list):
            for ind, input_name in enumerate(inputs):
                if not self._backend.has_variable(input_name):
                    input_name = self._add_label(input_name)
                    inputs[ind] = input_name
                    # assert self._backend.has_variable(input_name)
        elif isinstance(inputs, str):
            if not self._backend.has_variable(inputs):
                inputs = self._add_label(inputs)
                inputs = [inputs]
                # assert self._backend.has_variable(inputs[-1])
        else:
            raise ValueError("the preprocessing of op_to_backend do not support this input type")
        if isinstance(outputs, list):
            for ind, output_name in enumerate(outputs):
                if not self._backend.has_variable(output_name):
                    output_name = self._add_label(output_name)
                    outputs[ind] = output_name
                    # assert self._backend.has_variable(output_name)
        elif isinstance(outputs, str):
            if not self._backend.has_variable(outputs):
                outputs = self._add_label(outputs)
                outputs = [outputs]
                # assert self._backend.has_variable(outputs[-1])
        else:
            raise ValueError("the preprocessing of op_to_backend do not support this input type")


        addcode_op = Op(outputs, func, inputs, owner=self, operation_type='_operations')
        self._backend.add_operation(addcode_op)

    def init_op_to_backend(self, outputs, func, inputs, prefer_device=0):
        addcode_op = Op(outputs, func, inputs, place=prefer_device, owner=self, operation_type='_operations')
        self._backend.register_initial(addcode_op)

    # adding label of the module object, cut from neurongroup and generalized to all Modules
    def _add_label(self, key):
        if isinstance(key, str):
            if key == '[dt]':
                return key
            elif '[updated]' in key:
                return self.id + ':' + '{' + key.replace('[updated]', "") + '}' + '[updated]'
            else:
                return self.id + ':' + '{' + key + '}'
        elif isinstance(key, VariableAgent):
            return key.var_name
        else:
            raise ValueError(" the key data type is not supported for add_label")

    def get_full_name(self, name):
        name = '{'+name+'}'
        full_name = None
        for key in self._var_names:
            if name in key:
                if full_name is not None:
                    raise ValueError("multiple variable with same name in this module")
                else:
                    full_name = key
        return full_name

    def get_value(self, name):
        full_name = self.get_full_name(name)
        if full_name is None:
            raise  ValueError("No such variable name in this module")
        else:
            return self._var_dict[full_name].value

    def set_value(self, name, value):
        name = '{' + name + '}'
        full_name = None
        for key in self._var_names:
            if name in key:
                if full_name is not None:
                    raise ValueError("multiple variable with same name in this module")
                else:
                    full_name = key
        if full_name is None:
            raise ValueError("No such variable name in this module")
        else:
            self._var_dict[full_name].value = value

    def _direct_set_variable(self, name, variable):
        # only for debug at the beginning of the network run
        name = '{' + name + '}'
        full_name = None
        for key in self._var_names:
            if name in key:
                if full_name is not None:
                    raise ValueError("multiple variable with same name in this module")
                else:
                    full_name = key
        if full_name is None:
            raise ValueError("No such variable name in this module")
        else:
            is_parameter = self._var_dict[full_name]._is_parameter
            if is_parameter:
                self._backend._parameters_dict[full_name] = variable
            else:
                self._backend._InitVariables_dict[full_name] = variable





class VariableAgent(object):
    def __init__(self, backend, var_name, is_parameter=False, dict_label=None):
        super(VariableAgent, self).__init__()
        assert isinstance(backend, spaic.Backend)
        self._backend: spaic.Backend = backend
        self._var_name = var_name
        self._is_parameter = is_parameter
        self.data_type = None
        self.device = None
        self.dict_label = dict_label

        self.set_funcs = []
        self.get_funcs = []

    @property
    def var_name(self):
        return self._var_name

    def new_labeled_agent(self, dict_label):
        assert (dict_label=='variables_dict' or dict_label=='update_dict'
                or dict_label=='reduce_dict' or dict_label=='temp_dict')
        agent = copy(self)
        agent.dict_label = dict_label
        return agent

    @property
    def value(self):
        if self.dict_label is None:
            return self._backend.get_varialble(self._var_name)
        elif self.dict_label == 'variables_dict':
            return self._backend._variables[self._var_name]
        elif self.dict_label == 'update_dict':
            return self._backend._update_dict[self._var_name]
        elif self.dict_label == 'reduce_dict':
            return self._backend._reduce_dict[self._var_name]
        elif self.dict_label == 'temp_dict':
            return self._backend._temp_dict[self._var_name]
        else:
            raise ValueError("can't find variable %s"%self._var_name)

    @value.setter
    def value(self, value):
        if self.dict_label is None:
            self._backend.set_variable_value(self._var_name, value, self._is_parameter)
        elif self.dict_label == 'update_dict':
            self._backend._update_dict[self._var_name] = value
        elif self.dict_label == 'reduce_dict':
            if self._var_name in self._backend._reduce_dict:
                self._backend._reduce_dict[self._var_name].append(value)
            else:
                self._backend._reduce_dict[self._var_name] = [value]
        elif self.dict_label == 'temp_dict':
             self._backend._temp_dict[self._var_name] = value
        elif self.dict_label == 'variables_dict':
            self._backend._variables[self._var_name] = value
        else:
            raise ValueError("can't set value of variable %s" % self._var_name)







class OperationCommand(object):
    def __init__(self, front_module, output, function, input):
        super(OperationCommand, self).__init__()
        assert isinstance(front_module, BaseModule)
        assert isinstance(output, list)
        assert isinstance(function, str) or callable(function)
        assert isinstance(input, list)

        self.front_module = front_module
        self.output = output
        self.function = function
        self.input = input
        self.training_only = False

    @property
    def enabled(self):
        if self.training_only:
            return self.front_module.enabled and self.front_module.training
        else:
            return self.front_module.enabled

@dataclass
class Op:
    '''
    Operation data class.
    '''
    output: Optional[List] = field(default_factory=list)
    func_name: Optional[str] = None
    input: Optional[List] = field(default_factory=list)

    place: Optional[Any] = None
    owner: Optional[BaseModule] = None
    requires_grad: Optional[bool] = False
    operation_type : Optional[str] = None # _opertaions, _init_operations, _standalone_operations
    func: Optional[Any] = None
    
    def set_identifier(self, nid=None):
        """Initialize self._identifier"""
        if nid is None:
            self._identifier = str(uuid.uuid1())
        else:
            self._identifier = nid
# class NetModule(BaseModule):
#     '''
#     Base class for snn network modules: assemblies, connection
#     '''
#
#     def __init__(self):
#         super(NetModule, self).__init__()
#
#         self.trainable_parameter_names = OrderedDict()
#
#     def add_trainable_names(self, name):
#         pass





# -*- coding: utf-8 -*-
"""
Created on 2020/8/17
@project: SPAIC
@filename: Library
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description: 
"""
from spaic.Network.Assembly import BaseModule
from spaic.Network.Network import Network


class NetLibrary(BaseModule):

    def __init__(self):
        super(NetLibrary, self).__init__()

        pass




# Copyright 2014 Jason Heeris, jason.heeris@gmail.com
# 
# This file is part of the gammatone toolkit, and is licensed under the 3-clause
# BSD license: https://github.com/detly/gammatone/blob/master/COPYING
from __future__ import division
import numpy as np

from .filters import make_erb_filters, centre_freqs, erb_filterbank

"""
This module contains functions for rendering "spectrograms" which use gammatone
filterbanks instead of Fourier transforms.
"""


def round_half_away_from_zero(num):
    """ Implement the round-half-away-from-zero rule, where fractional parts of
    0.5 result in rounding up to the nearest positive integer for positive
    numbers, and down to the nearest negative number for negative integers.
    """
    return np.sign(num) * np.floor(np.abs(num) + 0.5)


def gtgram_strides(fs, window_time, hop_time, filterbank_cols):
    """
    Calculates the window size for a gammatonegram.
    
    @return a tuple of (window_size, hop_samples, output_columns)
    """
    nwin        = int(round_half_away_from_zero(window_time * fs))
    hop_samples = int(round_half_away_from_zero(hop_time * fs))
    columns     = (1
                    + int(
                        np.floor(
                            (filterbank_cols - nwin)
                            / hop_samples
                        )
                    )
                  )
        
    return (nwin, hop_samples, columns)


def gtgram_xe(wave, fs, channels, f_min):
    """ Calculate the intermediate ERB filterbank processed matrix """
    cfs = centre_freqs(fs, channels, f_min)
    fcoefs = np.flipud(make_erb_filters(fs, cfs))
    xf = erb_filterbank(wave, fcoefs)
    xe = np.power(xf, 2)
    return xe


def gtgram(
    wave,
    fs,
    window_time, hop_time,
    channels,
    f_min):
    """
    Calculate a spectrogram-like time frequency magnitude array based on
    gammatone subband filters. The waveform ``wave`` (at sample rate ``fs``) is
    passed through an multi-channel gammatone auditory model filterbank, with
    lowest frequency ``f_min`` and highest frequency ``f_max``. The outputs of
    each band then have their energy integrated over windows of ``window_time``
    seconds, advancing by ``hop_time`` secs for successive columns. These
    magnitudes are returned as a nonnegative real matrix with ``channels`` rows.
    
    | 2009-02-23 Dan Ellis dpwe@ee.columbia.edu
    |
    | (c) 2013 Jason Heeris (Python implementation)
    """
    xe = gtgram_xe(wave, fs, channels, f_min)    
    
    nwin, hop_samples, ncols = gtgram_strides(
        fs,
        window_time,
        hop_time,
        xe.shape[1]
    )
    
    y = np.zeros((channels, ncols))
    
    for cnum in range(ncols):
        segment = xe[:, cnum * hop_samples + np.arange(nwin)]
        y[:, cnum] = np.sqrt(segment.mean(1))
    
    return y


# -*- coding: utf-8 -*-
"""
Created on 2021/6/1
@project: SPAIC
@filename: Rate_Modulation
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description: 
"""

from spaic.Network.BaseModule import Op
from .Learner import Learner
from spaic import Backend
import torch

class Rate_Modulate(Learner):
    '''
        neuron spiking rate modulation
    '''
    def __init__(self,trainable=None, **kwargs): #mean_rate, modulate_effect, ave_rate, trainable=None, post_var_name='O'):
        
        super(Rate_Modulate, self).__init__(trainable)
        self.m_r = kwargs.get("m_r")
        self.alpha = kwargs.get("alpha",0.001)
        # self.beta = kwargs.get("beta",0.9)
        self.trainable = trainable
        self.prefered_backend = ['pytorch']
        self.name = 'RateMode'
        self.post_var_name = kwargs.get("post_var_name", 'O')

        self.modulations = []


    def build(self, backend: Backend):
        super(Rate_Modulate, self).build(backend)
        ind = 0
        for key, con in self.trainable_connections.items():
            for unit_con in con.unit_connections:
                pre_group = unit_con[0]
                post_group = unit_con[1]
                weight = backend._parameters_dict[con.get_weight_name(pre_group, post_group)]
                if isinstance(self.m_r, list):
                    m_r = self.m_r[ind]
                else:
                    m_r = self.m_r
                if isinstance(self.alpha, list):
                    alpha = self.alpha[ind]
                else:
                    alpha = self.alpha
                if isinstance(self.post_var_name, list):
                    post_var_name = self.post_var_name[ind]
                else:
                    post_var_name = self.post_var_name

                output_name = post_group.id + ':' + '{' + post_var_name + '}'
                method = Modulation_method(weight, m_r, alpha, output_name)
                self.modulations.append(method)
                self.op_to_backend(None, method.record_rate, output_name)
                self.init_op_to_backend(None, method.modulate_step, [])
            ind += 1

    @property
    def rate_loss(self):
        loss = 0
        for mod in self.modulations:
            if mod.rate_loss is None:
                return None
            else:
                loss = loss + mod.rate_loss
        return loss



Learner.register("rate_modulate", Rate_Modulate)



class Modulation_method():

    def __init__(self, weight,  m_r, alpha, var_name):
        super(Modulation_method, self).__init__()
        self.var_name = var_name
        self.rec_rate = None
        self.weight = weight
        self.m_r = m_r
        self.alpha = alpha
        self.init_rate = torch.zeros([1, weight.shape[0]], device=weight.device)

    def record_rate(self, out):
        if '{[2]' in self.var_name:
            self.rec_rate = self.rec_rate + out[:,0, ...]
        else:
            self.rec_rate = self.rec_rate + out


    def modulate_step(self):
        self.rec_rate = self.init_rate

    @property
    def rate_loss(self):
        if self.rec_rate is not None:
            rate_loss = self.alpha*torch.norm(self.rec_rate - self.m_r)
            return rate_loss
        else:
            return None














# -*- coding: utf-8 -*-
"""
Created on 2020/8/12
@project: SPAIC
@filename: STCA_music_network
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description:
建立网络的例子
"""
import spaic as sf
from torch import nn
# nn.Module
class STCANet(sf.Network):

    def __init__(self):
        super(STCANet, self).__init__()

        self.input = sf.Node(100)
        self.layer1 = sf.NeuronGroup(500, 'default')
        self.layer2 = sf.NeuronGroup(10, 'default')
        self.output = sf.Node(10)

        self.connect(self.input, self.layer1)
        self.link2 = sf.Connection(self.layer1, self.layer2, 'full')
        self.connect(self.layer2, self.output, 'one_hot')



if __name__ == "__main__":
    # a = sf.Network()
    a = STCANet()
    # a.layer1 = sf.Node(100)

    print(a.get_groups())



# -*- coding: utf-8 -*-
"""
Created on 2020/8/12
@project: SPAIC
@filename: Learner
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description:
定义学习模块，包括各种学习算法对仿真计算过程中插入的各种计算模块，以及记录需要学习连接的接口
"""
# from spaic.Network.Connection import Connection
# from spaic.Neuron.Neuron import NeuronGroup
from spaic.Network.Assembly import BaseModule
from abc import ABC, abstractmethod
from collections import OrderedDict
import torch
import spaic
import torch.nn.functional as F
import numpy as np

class Learner(BaseModule, ABC):

    '''
        Base learner model for all the learner model.

        Args:
            parameters(dict) : The parameters for learner.
            super_parameters(dict) : Super parameters for future use.
            backend_functions(dict) : Contains all the learner model we can choose.

            name(str) : The typical name for the learner.
            preferred_backend(str) : Choose which kind of backend to use. Like "Pytorch", "Tensorflow" or "Jax".
            trainable_groups(dict) : Trainable container, includes nodes and layers to train.
            trainable_connections(dict) : Trainable container, includes connections to train.

            init_trainable: The initial state of this learner of whether it is trainable.

        Methods:
            add_trainable(self, trainable) : Add target object (Network, Assembly, Connection,
                            or list of them) to the trainable container
            build(self, backend) : Build Learner, choose the backend as user wish, if we have already finished the api.

    '''
    _class_label = '<learner>'
    learning_algorithms = dict()
    learning_optims = dict()
    optim_dict = {'Adam': torch.optim.Adam, 'AdamW': torch.optim.AdamW,
                  'SparseAdam': torch.optim.SparseAdam, 'Adamax': torch.optim.Adamax,
                  'ASGD': torch.optim.ASGD, 'LBFGS': torch.optim.LBFGS,
                  'RMSprop': torch.optim.RMSprop, 'Rpop': torch.optim.Rprop,
                  'SGD': torch.optim.SGD, 'Adadelta': torch.optim.Adadelta,
                  'Adagrad': torch.optim.Adagrad}

    lr_schedule_dict = {'LambdaLR': torch.optim.lr_scheduler.LambdaLR,
                        'StepLR': torch.optim.lr_scheduler.StepLR,
                        'MultiStepLR': torch.optim.lr_scheduler.MultiStepLR,
                        'ExponentialLR': torch.optim.lr_scheduler.ExponentialLR,
                        'CosineAnnealingLR': torch.optim.lr_scheduler.CosineAnnealingLR,
                        'ReduceLROnPlateau': torch.optim.lr_scheduler.ReduceLROnPlateau,
                        'CyclicLR': torch.optim.lr_scheduler.CyclicLR,
                        'CosineAnnealingWarmRestarts': torch.optim.lr_scheduler.CosineAnnealingWarmRestarts}

    def __init__(self, trainable=None, pathway=None, algorithm=('STCA', 'STBP', 'RSTDP', '...'), name=None, **kwargs):

        super(Learner, self).__init__()

        self.parameters = kwargs
        self.super_parameters = OrderedDict()
        self.backend_functions = OrderedDict()

        self.name = self.set_name(name)
        self.optim_name = None
        self.optim = None
        self.lr_schedule_name = None
        self.gradient_based = True
        self.prefered_backend = ['pytorch']
        self.trainable_groups = OrderedDict()
        self.trainable_connections = OrderedDict()
        self.trainable_nodes = OrderedDict()
        self.trainable_modules = OrderedDict()
        self.trainable_others = OrderedDict()
        self.init_trainable = trainable
        self.training_param_name = []
        self.param_run_update = kwargs.get("param_run_update", False)
        self._variables = dict()
        self._constant_variables = dict()
        self._tau_constant_variables = dict()
        self._tau_membrane_variables = dict()


        self.pathway_groups = OrderedDict()
        self.pathway_connections = OrderedDict()
        self.pathway_nodes = OrderedDict()
        self.pathway_modules = OrderedDict()
        self.pathway_others = OrderedDict()
        self.init_pathway = pathway
        self._operations = []
        self._active = True


    def add_trainable(self, trainable: list):
        '''
            Add target object (Assembly, Connection, or list of them) to the trainable container
            Args:
                trainable(list) : The trainable target waiting for added.
        '''
        from spaic.Network.Assembly import Assembly
        from spaic.Network.Connections import Connection
        from spaic.Neuron.Neuron import NeuronGroup
        from spaic.Neuron.Module import Module
        from spaic.Neuron.Node import Node

        if not isinstance(trainable, list):
            trainable = [trainable]

        for target in trainable:
            if isinstance(target, NeuronGroup):
                self.trainable_groups[target.id] = target
            elif isinstance(target, Connection):
                self.trainable_connections[target.id] = target
            elif isinstance(target, Node):
                self.trainable_nodes[target.id] = target
            elif isinstance(target, Module):
                self.trainable_modules[target.id] = target
            elif isinstance(target, Assembly):
                for sub_t in target.get_groups():
                    trainable.append(sub_t)
                for sub_t in target.get_connections():
                    trainable.append(sub_t)
            elif isinstance(target, BaseModule):
                self.trainable_others[target.id] = target


    def add_pathway(self, pathway: list):
        '''
            Add target object (Assembly, Connection, or list of them) to the pathway container
            Args:
                pathway(list) : The pathway target waiting for added.
        '''
        from spaic.Network.Assembly import Assembly
        from spaic.Network.Connections import Connection
        from spaic.Neuron.Neuron import NeuronGroup
        from spaic.Neuron.Module import Module
        from spaic.Neuron.Node import Node

        if not isinstance(pathway, list):
            pathway = [pathway]

        for target in pathway:
            if isinstance(target, NeuronGroup):
                self.pathway_groups[target.id] = target
            elif isinstance(target, Connection):
                self.pathway_connections[target.id] = target
            elif isinstance(target, Node):
                self.pathway_nodes[target.id] = target
            elif isinstance(target, Module):
                self.pathway_modules[target.id] = target
            elif isinstance(target, Assembly):
                for sub_t in target.get_groups():
                    pathway.append(sub_t)
                for sub_t in target.get_connections():
                    pathway.append(sub_t)
            elif isinstance(target, BaseModule):
                self.pathway_others[target.id] = target

    def build(self, backend: spaic.Backend):
        '''
            Build Learner, choose the backend as user wish, if we have already finished the api.
            Args:
                backend(backend) : Backend we have.
        '''
        if self.init_trainable is not None:  # If user has given the 'trainable' parameter.
            self.add_trainable(self.init_trainable)
        if self.init_pathway is not None:
            self.add_pathway(self.init_pathway)

        if backend.backend_name in self.prefered_backend:
            self._backend = backend
        else:
            raise ValueError(
                "the backend %s is not supported by the learning rule %s" % (backend.backend_name, self.name))
        if self.optim_name is not None:
            self.build_optimizer()
        else:
            self.get_param()

        if self.lr_schedule_name is not None:
            self.build_lr_schedule()

        # set all trainable or pathway compontents to requires_grad=Ture
        if self.gradient_based:
            for node in self.trainable_nodes.values():
                for op in node._ops:
                    op.requires_grad = True
            for group in self.trainable_groups.values():
                for op in group._ops:
                    op.requires_grad = True
            for con in self.trainable_connections.values():
                for op in con._ops:
                    op.requires_grad = True
            for other in self.trainable_others.values():
                for op in other._ops:
                    op.requires_grad = True
            for mod in self.trainable_modules.values():
                mod.module.requires_grad_()
                for op in mod._ops:
                    op.requires_grad = True

            for node in self.pathway_nodes.values():
                for op in node._ops:
                    op.requires_grad = True
            for group in self.pathway_groups.values():
                for op in group._ops:
                    op.requires_grad = True
            for con in self.pathway_connections.values():
                for op in con._ops:
                    op.requires_grad = True
            for other in self.pathway_others.values():
                for op in other._ops:
                    op.requires_grad = True
            for mod in self.pathway_modules.values():
                mod.module.requires_grad_()
                for op in mod._ops:
                    op.requires_grad = True


        # add learner variables to the backend
        self.dt = backend.dt
        for (key, tau_var) in self._tau_constant_variables.items():
            tau_var = np.exp(-self.dt / tau_var)
            shape = ()
            self.variable_to_backend(self._add_label(key), shape, value=tau_var)
        for (key, tau_membrane_var) in self._tau_membrane_variables.items():
            tau_membrane_var = self.dt/tau_membrane_var
            shape = ()  # (1, neuron_num)
            self.variable_to_backend(self._add_label(key), shape, value=tau_membrane_var)
        for (key, var) in self._constant_variables.items():
            if isinstance(var, np.ndarray):
                if var.size > 1:
                    var_shape = var.shape
                    shape = (1, *var_shape)  # (1, shape)
                else:
                    shape = ()
            elif isinstance(var, list):
                if len(var) > 1:
                    var_len = len(var)
                    shape = (1, var_len)  # (1, shape)
                else:
                    shape = ()
            else:
                shape = ()
            self.variable_to_backend(self._add_label(key), shape, value=var)
        for (key, var) in self._variables.items():
            if isinstance(var, np.ndarray):
                if var.size > 1:
                    var_shape = var.shape
                    shape = (1, *var_shape)  # (1, shape)
                else:
                    shape = ()
            elif isinstance(var, list):
                if len(var) > 1:
                    var_len = len(var)
                    shape = (1, var_len)  # (1, shape)
                else:
                    shape = ()
            else:
                shape = ()
            self.variable_to_backend(self._add_label(key), shape, value=var)

        # build custom learning rules
        if self.is_overridden(self.custom_rule):
            self.custom_rule(backend)
        if self.is_overridden(self.connection_rule):
            for conn in self.trainable_connections.values():
                self.connection_rule(conn, backend, 'trainable')
            for conn in self.pathway_connections.values():
                self.connection_rule(conn, backend, 'pathway')
        if self.is_overridden(self.neuron_rule):
            for neuron in self.trainable_groups.values():
                self.neuron_rule(neuron, backend, 'trainable')
            for neuron in self.pathway_groups.values():
                self.neuron_rule(neuron, backend, 'pathway')



    def __new__(cls, trainable=None, pathway=None, algorithm=('STCA', 'STBP', 'RSTDP', '...'), **kwargs):
        if cls is not Learner:
            return super().__new__(cls)

        if algorithm == ('STCA', 'STBP', 'RSTDP', '...'):
            return super().__new__(cls)
        elif algorithm.lower() in cls.learning_algorithms:
            return cls.learning_algorithms[algorithm.lower()](trainable=trainable, pathway=None, **kwargs)
        else:
            raise ValueError("No algorithm %s in algorithm list" % algorithm)

    def active(self):
        self._active = True
        #TODO: setting all learnerable and pathway into gradient = True
    def deactive(self):
        self._active = False
        #TODO: setting all learnerable and pathway into gradient = Fasle

    def set_optimizer(self, optim_name, optim_lr, **kwargs):
        self.optim_lr = optim_lr
        self.optim_para = kwargs
        self.optim_name = optim_name

        if self.optim_name not in Learner.optim_dict.keys():
            raise ValueError("No optim %s in optim list" % Learner.optim_dict)

    def set_schedule(self, lr_schedule_name, **kwargs):

        self.lr_schedule_para = kwargs

        self.lr_schedule_name = lr_schedule_name

        if self.lr_schedule_name not in Learner.lr_schedule_dict.keys():
            raise ValueError("No lr_schedule %s in lr_schedule list")

    def get_param(self):
        param = list()
        var_name = list()
        for key, conn in self.trainable_connections.items():
            for name in conn._var_names:
                var_name.append(name)
        for key, node in self.trainable_nodes.items():
            for name in node._var_names:
                var_name.append(name)
        for key, group in self.trainable_groups.items():
            for name in group._var_names:
                var_name.append(name)
        self.training_param_name = []
        for key, value in self._backend._parameters_dict.items():
            if key in var_name:
                value.requires_grad = True
                param.append(value)
                self.training_param_name.append(key)
            else:
                value.requires_grad =False


        for mod in self.trainable_modules.values():
            param.extend(mod.parameters)

        return param

    def get_varname(self, key):
        name = self.name + ':{' + key + '}'
        return name
    def get_var_names(self):
        return self._var_names

    def build_optimizer(self):

        param = self.get_param()
        self.optim = Learner.optim_dict[self.optim_name](param, self.optim_lr, **self.optim_para)

    def build_lr_schedule(self):

        self.schedule = Learner.lr_schedule_dict[self.lr_schedule_name](self.optim, **self.lr_schedule_para)

    def optim_step(self):
        if self.param_run_update:
            with torch.no_grad():
                for key, value in self._backend._parameters_dict.items():
                    if key in self.training_param_name:
                        varialbe_value = self._backend._variables[key]
                        if varialbe_value is not value:
                            value.data = varialbe_value.data
                            self._backend._variables[key] = value #Is this step needed?

        if self.optim is not None:
            self.optim.step()

    def optim_zero_grad(self):

        self.optim.zero_grad()

    def optim_schedule(self):

        self.schedule.step()

    @staticmethod
    def register(name, algorithm):
        name = name.lower()
        if name in Learner.learning_algorithms:
            raise ValueError(('A learning algorithm with the name "%s" has already been registered') % name)

        if not issubclass(algorithm, Learner):
            raise ValueError(
                ('Given algorithm of type %s does not seem to be a valid algorithm.' % str(type(algorithm))))

        Learner.learning_algorithms[name] = algorithm

    def custom_rule(self, backend: spaic.Backend):
        return None

    def connection_rule(self, con : spaic.Connection, backend: spaic.Backend, obj_type='trainable'):
        return None


    def neuron_rule(self, neuron: spaic.NeuronGroup, backend: spaic.Backend, obj_type='trainable'):
        return None


    def is_overridden(self, func):
        if isinstance(self, Learner):
            return False
        super_func = getattr(super(type(self), self), func.__name__)
        return super_func != func



class ReSuMe(Learner):

    def __init__(self):
        super(ReSuMe, self).__init__()
        pass


# Learner.register("ReSuMe", ReSuMe)


class FORCE(Learner):

    def __init__(self):
        super(FORCE, self).__init__()
        pass


# Learner.register("force", FORCE)



# -*- coding: utf-8 -*-
"""
Created on 2020/8/12
@project: SPAIC
@filename: Monitor
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description:
定义神经集群放电以及神经元状态量、连接状态量的仿真记录模块
"""
from spaic.Network.Assembly import BaseModule, Assembly
from spaic.Network.Connections import Connection
from spaic.Learning.Learner import Learner
from spaic.Backend.Backend import Backend
import numpy as np
import torch
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.image import AxesImage
from mpl_toolkits.axes_grid1 import make_axes_locatable


class Monitor(BaseModule):

    def __init__(self, target, var_name, index='full', dt=None, get_grad=False, nbatch=True):
        super().__init__()
        if isinstance(target, Assembly):
            self.target = target
            self.target_type = 'Assembly'
        elif isinstance(target, Connection):
            self.target = target
            self.target_type = 'Connection'
        elif isinstance(target, Learner):
            self.target = target
            self.target_type = 'Learner'
        elif target == None:
            self.target = None
            self.target_type = None
        else:
            raise ValueError("The target does not belong to types that can be watched (Assembly, Connection).")

        self.var_name = '{'+var_name+'}'
        self.index = index
        self.var_container = None
        self.get_grad = get_grad
        self.nbatch = nbatch
        self._nbatch_records = []      # all time window's record
        self._nbatch_times = []
        self._records = []   # single time window's record
        self._times = []
        self.dt = dt
        self.is_recording = True
        self.new_record = True

    def check_var_name(self, var_name):
        '''
        Check if variable is in the target model, and add the target id label to the variable name.

        Parameters
        ----------
        var_name : original variable name

        Returns : modified variable name
        -------

        '''
        tar_var_name = None
        if var_name[1:-1] in self.backend._variables.keys():
            tar_var_name = var_name[1:-1]
        else:
            for tar_name in self.target.get_var_names():    # 没有中间变量
                if var_name in tar_name:
                    tar_var_name = tar_name
                    break

        if tar_var_name is not None:
            return tar_var_name
        else:
            raise ValueError(" Variable %s is not in the target model"%var_name)

    def get_str(self, level):
        pass
    def monitor_on(self):
        self.is_recording = True

    def monitor_off(self):
        self.is_recording = False

    def clear(self):
        NotImplementedError()

    def build(self, backend: Backend):
        NotImplementedError()


    def init_record(self):
        NotImplementedError()

    def update_step(self):
        NotImplementedError()

    def push_data(self, data, time):
        "push data to monitor by backend"
        self._records.append(data)
        self._times.append(time)




class SpikeMonitor(Monitor):
    def __init__(self, target, var_name='O', index='full', dt=None, get_grad=False, nbatch=False):
        super().__init__(target=target, var_name=var_name, index=index, dt=dt, get_grad=get_grad, nbatch=nbatch)
        self._transform_len = 0
        self._nbatch_index = []      # all time window's record
        self._nbatch_times = []
        self._spk_index = []
        self._spk_times = []
        self._records = []   # single time window's record
        self._times = []


    def build(self, backend: Backend):
        self.backend = backend
        self.backend._monitors.append(self)
        self.var_name = self.check_var_name(self.var_name)
        self.shape = self.backend._variables[self.var_name].shape
        if self.dt is None:
            self.dt = self.backend.dt

    def clear(self):
        self._transform_len = -1
        self._nbatch_index = []      # all time window's record
        self._nbatch_times = []
        self._spk_index = []
        self._spk_times = []
        self._records = []   # single time window's record
        self._times = []


    def init_record(self):
        self.new_record = True
        if len(self._spk_index) > 0:
            if self.nbatch is True:
                if isinstance(self._spk_index[0], torch.Tensor):
                    self._nbatch_index.append(torch.stack(self._spk_index[1:], dim=-1).cpu().detach().numpy())
                else:
                    self._nbatch_index.append(np.stack(self._spk_index[1:], axis=-1))
                self._nbatch_times.append(self._times[1:])
            elif self.nbatch > 0:
                if isinstance(self._spk_index[0], torch.Tensor):
                    self._nbatch_index.append(torch.stack(self._spk_index[1:], dim=-1).cpu().detach().numpy())
                else:
                    self._nbatch_index.append(np.stack(self._spk_index[1:], axis=-1))
                self._nbatch_times.append(self._times[1:])
                if len(self._nbatch_times) > self.nbatch:
                    self._nbatch_index = self._nbatch_index[-self.nbatch:]
                    self._nbatch_times = self._nbatch_times[-self.nbatch:]

        self._records = []   # single time window's record
        self._times = []
        self._transform_len = -1

    def push_spike_train(self, spk_times, spk_index, batch_index=0):
        if len(self._spk_index) < batch_index+1:
            add_num = batch_index + 1 - len(self._spk_index)
            for _ in range(add_num):
                self._spk_index.append([])
                self._spk_times.append([])
        if isinstance(spk_times, list) or isinstance(spk_times, tuple):
            self._spk_times[batch_index].extend(spk_times)
            self._spk_index[batch_index].extend(spk_index)
        else:
            self._spk_times[batch_index].append(spk_times)
            self._spk_index[batch_index].append(spk_index)

        #to override the _spike_transform function when getting spk_times and spk_index
        self._transform_len = 1


    def update_step(self, variables):
        '''
        Recoding the variable values of the current step.

        Returns
        -------

        '''
        if self.is_recording is False:
            return

        from decimal import Decimal
        acttime = Decimal(self.backend.time/ self.dt).quantize(Decimal(str(min(self.dt, 0.1))), rounding="ROUND_HALF_UP")
        if int(10000 * float(acttime)) % 10000 == 0:
            record_value = variables[self.var_name]
            if self.get_grad:
                variables[self.var_name].retain_grad()
            if self.index == 'full':
                self._records.append(record_value)
                self._times.append(self.backend.time)
            else:
                if len(self.index) == record_value.ndim:
                    self._records.append(record_value[self.index])
                    self._times.append(self.backend.time)
                else:
                    assert len(self.index) == record_value.ndim -1
                    if self.backend.backend_name == 'pytorch':
                        record_value = torch.movedim(record_value, 0, -1)
                        indexed_value = record_value[tuple(self.index)]
                        indexed_value = torch.movedim(indexed_value, -1, 0)
                    else:
                        record_value = np.array(record_value)
                        record_value = np.moveaxis(record_value, 0, -1)
                        indexed_value = record_value[tuple(self.index)]
                        indexed_value = np.moveaxis(indexed_value, -1, 0)
                    self._records.append(indexed_value)
                    self._times.append(self.backend.time)

    def _spike_transform(self):
        batch_size = self.backend.get_batch_size()
        if len(self._records) > self._transform_len:
            self._transform_len = len(self._records)
            self._spk_index = []
            self._spk_times = []
            # self._spk_p = []
            if isinstance(self._records[0], torch.Tensor):
                step = len(self._records)
                # for i,x in enumerate(self._records):
                #     self._records[i] = x.cpu()
                rec_spikes = torch.stack(self._records, dim=-1).detach()
                if '{[2]' in self.var_name:
                    for ii in range(batch_size):
                        rec_spikes_i = rec_spikes[ii,0,...].bool().reshape(-1).cpu()
                        rec_spikes_t = rec_spikes[ii,1,...].reshape(-1)
                        num = int(rec_spikes_i.size(0)/step)
                        time_seq = torch.tensor(self._times).unsqueeze(dim=0).expand(num, -1).reshape(-1)
                        indx_seq = torch.arange(0, num).unsqueeze(dim=1).expand(-1, step).reshape(-1)
                        time_seq = (torch.masked_select(time_seq - rec_spikes_t, rec_spikes_i) ).numpy()
                        indx_seq = torch.masked_select(indx_seq, rec_spikes_i).numpy()
                        self._spk_index.append(indx_seq)
                        self._spk_times.append(time_seq)

                else:
                    for ii in range(batch_size):
                        if rec_spikes.dtype.is_complex:
                            rec_spikes_i = (rec_spikes[ii, ...].imag.bool()*rec_spikes[ii, ...].real.gt(0.0)).reshape(-1).cpu()
                        else:
                            rec_spikes_i = rec_spikes[ii,...].bool().reshape(-1).cpu()

                        num = int(rec_spikes_i.size(0)/step)
                        time_seq = torch.tensor(self._times).unsqueeze(dim=0).expand(num, -1).reshape(-1)
                        indx_seq = torch.arange(0, num).unsqueeze(dim=1).expand(-1, step).reshape(-1)
                        time_seq = torch.masked_select(time_seq, rec_spikes_i).numpy()
                        indx_seq = torch.masked_select(indx_seq, rec_spikes_i).numpy()
                        self._spk_index.append(indx_seq)
                        self._spk_times.append(time_seq)
                        # self._spk_p.append(torch.masked_select(rec_spikes[ii, ...].real.reshape(-1).cpu(), rec_spikes_i).numpy())


    @property
    def spk_times(self):
        self._spike_transform()
        return self._spk_times

    @property
    def spk_index(self):
        self._spike_transform()
        return self._spk_index

    @property
    # def spk_p(self):
    #     self._spike_transform()
    #     return self._spk_p

    @property
    def spk_grad(self):
        pass
        return None

    @property
    def time_spk_rate(self):
        if isinstance(self._records[0], torch.Tensor):
            if '{[2]' in self.var_name:
                spike = torch.stack(self._records, dim=-1).cpu().detach()[:,0,...]
            else:
                spike = torch.stack(self._records, dim=-1).cpu().detach()
            return torch.mean(spike, dim=0).numpy()
        else:
            if '{[2]' in self.var_name:
                spike = np.stack(self._records, axis=-1)[:,0,...]
            else:
                spike = np.stack(self._records, axis=-1)
            return np.mean(spike, axis=0).numpy()

    @property
    def time_pop_rate(self):
        if isinstance(self._records[0], torch.Tensor):
            if '{[2]' in self.var_name:
                spike = torch.stack(self._records, dim=-1).cpu().detach()[:,0,...]
            else:
                spike = torch.stack(self._records, dim=-1).cpu().detach()
            return torch.mean(spike, dim=1).numpy()
        else:
            if '{[2]' in self.var_name:
                spike = np.stack(self._records, axis=-1)[:,0,...]
            else:
                spike = np.stack(self._records, axis=-1)
            return np.mean(spike, axis=1).numpy()

    # def smooth_pop_rate(self, window=1.0):
    #

    @property
    def spk_count(self):
        if isinstance(self._records[0], torch.Tensor):
            spike = torch.stack(self._records, dim=-1).cpu()
            if spike.dtype.is_complex:
                spike = spike.real
            return torch.sum(spike.gt(0.0), dim=-1).numpy()
        else:
            spike = np.stack(self._records, axis=-1).__gt__(0.0)
            return np.sum(spike, axis=-1)



    @property
    def time(self):
        return np.stack(self._times, axis=-1)









class StateMonitor(Monitor):

    def __init__(self, target, var_name, index='full', dt=None, get_grad=False, nbatch=False):
        # TODO: 初始化有点繁琐，需要知道record的变量，考虑采用更直接的监控函数
        super().__init__(target=target, var_name=var_name, index=index, dt=dt, get_grad=get_grad, nbatch=nbatch)

        self._nbatch_records = []      # all time window's record
        self._nbatch_times = []
        self._records = []   # single time window's record
        self._times = []



    def build(self, backend: Backend):
        self.backend = backend
        self.backend._monitors.append(self)
        self.var_name = self.check_var_name(self.var_name)
        if self.index != 'full':
            self.index = tuple(self.index)
        if self.dt is None:
            self.dt = self.backend.dt


    def clear(self):
        self._nbatch_records = []      # all time window's record
        self._nbatch_times = []
        self._records = []   # single time window's record
        self._times = []


    def init_record(self):
        '''
        Inite record of new trial
        Returns:

        '''
        self.new_record = True
        self._last_step_time = 0
        if len(self._records) > 0:
            if self.nbatch is True:
                if isinstance(self._records[0], torch.Tensor):
                    self._nbatch_records.append(torch.stack(self._records, dim=-1).cpu().detach().numpy())
                else:
                    self._nbatch_records.append(np.stack(self._records, axis=-1))
                self._nbatch_times.append(self._times)
            elif self.nbatch > 0:
                if isinstance(self._records[0], torch.Tensor):
                    self._nbatch_records.append(torch.stack(self._records, dim=-1).cpu().detach().numpy())
                else:
                    self._nbatch_records.append(np.stack(self._records, axis=-1))
                self._nbatch_times.append(self._times)
                if len(self._nbatch_times) > self.nbatch:
                    self._nbatch_records = self._nbatch_records[-self.nbatch:]
                    self._nbatch_times = self._nbatch_times[-self.nbatch:]


            self._records = []
            self._times = []



    def update_step(self, variables):
        '''
        Recoding the variable values of the current step.

        Returns
        -------

        '''
        if self.is_recording is False:
            return

        # only data in variable_dict can be recorded now
        from decimal import Decimal
        acttime = Decimal(self.backend.time/ self.dt).quantize(Decimal(str(min(self.dt, 0.1))), rounding="ROUND_HALF_UP")
        if int(10000 * float(acttime)) % 10000 == 0:

            record_value = variables[self.var_name]
            if self.get_grad:
                var = variables[self.var_name]
                if var.requires_grad is True:
                    var.retain_grad()
            if self.index == 'full':
                self._records.append(record_value)
                self._times.append(self.backend.time)
            else:
                if len(self.index) == record_value.ndim:
                    self._records.append(record_value[self.index])
                    self._times.append(self.backend.time)
                else:
                    assert len(self.index) == record_value.ndim -1
                    if self.backend.backend_name == 'pytorch':
                        record_value = torch.movedim(record_value, 0, -1)
                        indexed_value = record_value[tuple(self.index)]
                        indexed_value = torch.movedim(indexed_value, -1, 0)
                    else:
                        record_value = np.array(record_value)
                        record_value = np.moveaxis(record_value, 0, -1)
                        indexed_value = record_value[tuple(self.index)]
                        indexed_value = np.moveaxis(indexed_value, -1, 0)
                    self._records.append(indexed_value)
                    self._times.append(self.backend.time)

        else:
            print(self.backend.time)
        self._last_step_time = self.backend.time


    @property
    def nbatch_values(self):
        if self.new_record:
            self._nbatch_records_ = self._nbatch_records + [torch.stack(self._records, dim=-1).cpu().detach().numpy()]
            self._nbatch_times_ = self._nbatch_times + [self._times]
            self.new_record = False
        return np.array([np.stack(records, axis=-1) for records in self._nbatch_records_])

    @property
    def nbatch_times(self):
        if self.new_record:
            self._nbatch_records_ = self._nbatch_records + [torch.stack(self._records, dim=-1).cpu().detach().numpy()]
            self._nbatch_times_ = self._nbatch_times + [self._times]
            self.new_record = False
        return np.array([np.stack(times, axis=-1) for times in self._nbatch_times_])

    @property
    def values(self):
        # return np.concatenate(self._records)
        if isinstance(self._records[0], torch.Tensor):
            # for i,x in enumerate(self._records):
            #     self._records[i] = x.cpu()
            return torch.stack(self._records, dim=-1).cpu().detach().numpy()
        else:
            return np.stack(self._records, axis=-1)

    @property
    def tensor_values(self):
        assert isinstance(self._records[0], torch.Tensor)
        return torch.stack(self._records, dim=-1)

    @property
    def grads(self):
        if self.get_grad:
            grads = []
            for v in self._records:
                if v.grad is not None:
                    grads.append(v.grad.cpu().numpy())
                else:
                    grads.append(torch.zeros_like(v).cpu().numpy())
            grads = np.stack(grads[1:], axis=-1)
            return grads
        else:
            return None


    @property
    def times(self):
        if isinstance(self._times[0], torch.Tensor):
            return torch.stack(self._times, dim=-1).cpu().detach().numpy()
        else:
            return np.stack(self._times, axis=-1)


    def plot_weight(self, **kwargs):
        neuron_id = kwargs.get('neuron_id')
        time_id = kwargs.get('time_id')
        batch_id = kwargs.get('batch_id', None)
        new_shape = kwargs.get('new_shape')
        reshape = kwargs.get('reshape')
        axes = kwargs.get('Axes', None)
        ims = kwargs.get('AxesImage', None)
        n_sqrt = kwargs.get('n_sqrt', None)
        side = kwargs.get('side', None)
        figsize = kwargs.get('figsize', (5, 5))
        cmap = kwargs.get('camp', 'hot_r')
        wmin = kwargs.get('wmin', 0)
        wmax = kwargs.get('wmax', 128)
        im = kwargs.get('im', None)

        if batch_id == None:
            value = self.values[:, :, time_id]
            # value = self.simulator._variables[
            #     'autoname1<net>_connection1<con>:autoname1<net>_layer1<neg><-autoname1<net>_input<nod>:{weight}']
            # value = value.cpu().detach().numpy()

            if reshape:

                # value = value.reshape(2, 5, side, side)
                #
                # value = value.transpose(0, 2, 1, 3)
                # value = value.reshape(2*side, 5*side)
                value = value.reshape(n_sqrt, n_sqrt, side, side)

                value = value.transpose(0, 2, 1, 3)
                value = value.reshape(n_sqrt * side, n_sqrt * side)
                square_weights = value

            else:
                square_weights = value

        else:
            value = self.nbatch_values[batch_id, :, time_id, :]
            if reshape:
                value = value.reshape(n_sqrt, n_sqrt, side, side)

                value = value.transpose(0, 2, 1, 3)
                value = value.reshape(n_sqrt * side, n_sqrt * side)
                square_weights = value
            else:
                square_weights = value
        if not im:
            fig, ax = plt.subplots(figsize=figsize)

            im = ax.imshow(square_weights, cmap=cmap, vmin=wmin, vmax=wmax)
            div = make_axes_locatable(ax)
            cax = div.append_axes("right", size="5%", pad=0.05)

            ax.set_xticks(())
            ax.set_yticks(())
            ax.set_aspect("auto")

            plt.colorbar(im, cax=cax)
            fig.tight_layout()
        else:
            im.set_data(square_weights)
        plt.pause(0.1)
        return im




# -*- coding: utf-8 -*-
"""
Created on 2020/8/12
@project: SPAIC
@filename: Environment
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description:
定义强化学习的环境交互模块
"""
from abc import ABC, abstractmethod
from spaic.IO.utils import RGBtoGray, GraytoBinary, reshape
# import gym
import numpy as np

'''
# Examples: initialize the environment of CartPole-v1
import gym
environment = gym.make('CartPole-v0')
for i_episode in range(20):
    observation = environment.reset()
    for t in range(100):
        environment.render()
        print(observation)
        action = environment.action_space.sample()
        observation, reward, done, info = environment.step(action)
        if done:
            print("Episode finished after {} timesteps".format(t+1))
            break
environment.close()
'''

class BaseEnvironment(ABC):
    """
    Abstract environment class.
    """
    def __init__(self):
        pass

    @abstractmethod
    def step(self, action: int):
        """
        Abstract method for ``step()``.

        Args:
            action (int): action to take in environment.
        """
        pass

    @abstractmethod
    def reset(self):
        """
        Abstract method for ``reset()``.
        """
        pass

    @abstractmethod
    def render(self):
        """
        Abstract method for ``render()``.
        """
        pass

    @abstractmethod
    def seed(self, seed):
        """
        Abstract method for ``seed()``.
        """
        pass

    @abstractmethod
    def close(self):
        """
        Abstract method for ``close()``.
        """
        pass

class GymEnvironment(BaseEnvironment):
    """
    Wrapper the OpenAI ``gym`` environments.
    """

    def __init__(self, name: str,  **kwargs):
        """
        Initializes the environment wrapper. This class makes the
        assumption that the OpenAI ``gym`` environment will provide an image
        of format HxW as an observation.

        Args:
            name (str): The name of an OpenAI ``gym`` environment.
            encoding (str): The key of encoding class which is used to encode observations into spike trains.

        Attributes:
            max_prob (float): Maximum spiking probability.
            clip_rewards (bool): Whether or not to use ``np.sign`` of rewards.
            binary (bool): Whether to convert the image to binary
        """
        import gym
        self.name = name
        self.environmet = gym.make(name)
        self.action_space = self.environmet.action_space
        self.action_num = self.action_space.n

        self.shape = kwargs.get('shape', None)
        self.binary = kwargs.get('binary', False)
        self.gray = kwargs.get('binary', True)
        self.flatten = kwargs.get('flatten', True)

        # Keyword arguments.
        self.max_prob = kwargs.get('max_prob', 1.0)
        self.clip_rewards = kwargs.get('clip_rewards', True)

        self.episode_step_count = 0

        self.obs = None
        self.reward = None

        assert (
            0.0 < self.max_prob <= 1.0
        ), "Maximum spiking probability must be in (0, 1]."

    def step(self, action):
        """
        Wrapper around the OpenAI ``gym`` environment ``step()`` function.

        Args:
            action (int): Action to take in the environment.
        Returns:
             Observation, reward, done flag, and information dictionary.
        """
        # Call gym's environment step function.
        self.obs, self.reward, self.done, info = self.environmet.step(action)

        if self.clip_rewards:
            self.reward = np.sign(self.reward)
        """
        After encoding the shape of 1D observations will become [Time_step, batch_size, length].
        2D observations are mono images. They will be flatten into 1D.
        3D observations are color images that will be converted to grayscale images and then will be flatten into 1D.
        """
        if len(self.obs.shape) >= 3 and self.gray:
            self.obs = RGBtoGray(self.obs)

        if self.binary:
            self.obs = GraytoBinary(self.obs)

        if self.shape is not None:
            if self.shape != self.obs.shape:
                self.obs = reshape(self.obs, self.shape)

        # Flatten
        if len(self.obs.shape) >= 2 and self.flatten:
            self.obs = self.obs.flatten()

        # Add the raw observation from the gym environment into the info for display.
        info['gym_obs'] = self.obs

        self.episode_step_count += 1

        # Return converted observations and other information.
        return self.obs, self.reward, self.done, info

    def reset(self):

        """
        Wrapper around the OpenAI ``gym`` environment ``reset()`` function.

        :return: Observation from the environment.
        """
        # Call gym's environment reset function.
        self.obs = self.environmet.reset()

        if len(self.obs.shape) >= 3 and self.gray:
            self.obs = RGBtoGray(self.obs)

        if self.binary:
            self.obs = GraytoBinary(self.obs)

        if self.shape is not None:
            self.shape = tuple(self.shape)
            if self.shape != self.obs.shape:
                self.obs = reshape(self.obs, self.shape)

        # Flatten
        if len(self.obs.shape) >= 2 and self.flatten:
            self.obs = self.obs.flatten()

        self.episode_step_count = 0

        return self.obs

    def render(self, mode):

        """
        Wrapper around the OpenAI ``gym`` environment ``render()`` function.
        """
        return self.environmet.render(mode)

    def seed(self, seed):
        """
        Wrapper around the OpenAI ``gym`` environment ``render()`` function.
        """
        self.environmet.seed(seed)

    def close(self):
        """
        Wrapper around the OpenAI ``gym`` environment ``close()`` function.
        """
        self.environmet.close()